// $Id: ta-util.cc 3877 2011-03-25 23:29:55Z flaterco $

/*
    TA (Transparent Archivist) (C) 2005 David Flater.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// Save this preprocessor constant's value before mhash.h clobbers it.
extern char const * const ta_version = VERSION;


#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <mhash.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "ta-util.hh"
#include <unistd.h>
#include <zlib.h>


// Size of hashing buffer in bytes.
// 0x100000 = 1 MiB
#define hashbufsize 0x100000

// Block size for createFile when fill_random.
#define randbufsize 0x100000


// in hash_ID:  one of MHASH_SHA512, MHASH_MD5, etc.
//   (see list in /usr/include/mutils/mglobal.h)
// in filename:  name of a regular file (not a directory, not a symlink).
// out shash:  stringified hash of that file
//   (e.g., for SHA-512 it would be 128 characters).
// Throws FileReadException if there's a problem reading the file,
//   whether because it doesn't exist or whatever.
void getHash (hashid hash_ID, const Dstr &filename, Dstr &shash) throw (FileReadException) {
  {
    struct stat64 s;
    if (lstat64 (filename.aschar(), &s) == 0) {
      if (!S_ISREG (s.st_mode)) {
	fprintf (stderr, "getHash called on something not a regular file: %s\n",
		 filename.aschar());
	exit (-1);
      }
    } else {
      perror (filename.aschar()); // lstat failed
      throw FileReadException();
    }
  }

  size_t blockSize = mhash_get_block_size (hash_ID);
  assert (blockSize > 0 && blockSize <= 64);
  unsigned char hash[64];

  {
    FILE *fp = fopen64 (filename.aschar(), "r");
    if (!fp) {
      perror (filename.aschar());
      throw FileReadException();
    }

    MHASH td = mhash_init (hash_ID);
    if (td == MHASH_FAILED) {
      fprintf (stderr, "Mhash failed to initialize\n");
      exit (-1);
    }

    // The return value of mhash is undocumented, but in the sources
    // it is hard-coded to always return 0.

    unsigned char buffer[hashbufsize];
    size_t nread;
    while ((nread = fread (buffer, 1, hashbufsize, fp)))
      mhash (td, buffer, nread);
    if (ferror (fp)) {
      fprintf (stderr, "Error reading from %s\n", filename.aschar());
      fclose (fp);
      throw FileReadException();
    }
    fclose (fp);
    mhash_deinit (td, hash);
  }

  shash = "";
  char buf[3];
  for (unsigned i=0; i<blockSize; ++i) {
    sprintf (buf, "%.2X", hash[i]);
    shash += buf;
  }
}


// Execute an external command.
//
// outfname    Name of file to which to redirect stdout (optional).
//
// argv[0]     The command
// argv[1]...  Arguments
// argv[n]     The list of arguments MUST be terminated by a NULL pointer.
//
// If the command returns a nonzero exit status, this is thrown as an
// exception of type int.
//
// All other types of failures result in exit with a helpful error message.
void sys (const char *outfname, char *const argv[]) throw (int) {
  pid_t pid = fork();
  switch (pid) {
  case -1:
    // failure
    perror ("fork");
    exit (-1);
  case 0:
    // child process
    if (outfname) {
      if (!(freopen (outfname, "w", stdout))) {
        perror (outfname);
        exit (-1);
      }
    }
    execvp (argv[0], argv);
    perror (argv[0]);
    exit (-1);
  default:
    // parent process
    // I don't feel like sorting out race conditions so let's just
    // wait on it.
    int status;
    pid_t waitret = waitpid (pid, &status, 0);
    assert (waitret == pid);
    if (WIFSIGNALED(status)) {
      fprintf (stderr, "%s: died from signal\n", argv[0]);
      exit (-1);
    }
    // The following check may be redundant, but I wouldn't bet on it.
    if (!WIFEXITED(status)) {
      fprintf (stderr, "%s: did not exit normally\n", argv[0]);
      exit (-1);
    }
    if (WEXITSTATUS(status))
      throw (int)(WEXITSTATUS(status));
  }
}


void sys (char *const argv[]) throw (int) {
  sys (NULL, argv);
}


// Create a file of the specified size.
// If the file exists it is overwritten.
// All errors result in exit.
//
// If fill_random, file is filled with random data.
// Otherwise, it is created sparse if possible.

void createFile (const Dstr &filename, ta_t size, bool fill_random) {
  FILE *fp = fopen64 (filename.aschar(), "w");
  if (!fp) {
    perror (filename.aschar());
    exit (-1);
  }
  if (size) {
    if (fill_random) {
      printf ("Filling image with random background\n");
      FILE *frnd = fopen64 ("/dev/urandom", "r");
      assert (frnd);
      ta_t remaining = size;
      char buf[randbufsize];
      while (remaining) {
        size_t wsize = randbufsize;
        if (remaining < wsize)
          wsize = remaining;
        assert (fread (buf, 1, wsize, frnd) == wsize);
        if (fwrite (buf, 1, wsize, fp) != wsize) {
  	  fprintf (stderr, "%s: fwrite failed\n", filename.aschar());
	  exit (-1);
        }
        remaining -= wsize;
      }
      fclose (frnd);
    } else {
      if (fseeko64 (fp, size-1, SEEK_SET)) {
	Dstr msg (filename.aschar());
	msg += ": fseeko64 failed";
	perror (msg.aschar());
	exit (-1);
      }
      if (fputc (0, fp) == EOF) {
	fprintf (stderr, "%s: fputc failed\n", filename.aschar());
	exit (-1);
      }
    }
  }
  if (fclose (fp)) {
    perror (filename.aschar());
    exit (-1);
  }
}


// Given:  filename.
// Return:  name of corresponding file in wdpath.
void getTarget (const Dstr &filename, const Dstr &wdpath, Dstr &target) {
  assert (wdpath.back() == '/');
  assert (filename.length());
  target = wdpath;
  if (filename[0] == '/')
    target += filename.ascharfrom(1);
  else
    target += filename;
}


// Make sure that a directory name is nonnull and ends in a slash.
void slashterm (Dstr &dirname) {
  assert (dirname.length());
  if (dirname.back() != '/')
    dirname += '/';
}


// Eliminate "." references and complain about "..".
static void fixDotDirs (Dstr &fname) {
  while (fname[0] == '.' && fname[1] == '/')
    fname /= 2;
  while (fname.repstr ("/./", "/"));
  // Can't end up with nothing because then it would have been a directory.
  assert (fname.length() > 0);

  // Can't end with "/.." because then it would have been a directory.
  if ((fname[0] == '.' && fname[1] == '.' && fname[2] == '/') ||
      fname.contains ("/../")) {
    fprintf (stderr, "ERROR: reference to .. directory in %s\n"
	             "Please don't mention .. on the ta command line.\n",
                     fname.aschar());
    exit (-1);
  }
}


// Update a regular file's size entry per zlib estimate of compression.
// Based on zpipe.c example at http://zlib.net/zpipe.c, version 1.4
static void compressSize (FileStuff &fs) {
  ta_t compressedSize = 0;

  int ret, flush;
  z_stream strm;
  const size_t chunkSize = 1048576;
  uint8_t in[chunkSize], out[chunkSize];

  strm.zalloc = Z_NULL;
  strm.zfree = Z_NULL;
  strm.opaque = Z_NULL;
  ret = deflateInit (&strm, Z_DEFAULT_COMPRESSION);
  if (ret != Z_OK) {
    fprintf (stderr, "zlib failed to initialize; return = %d\n", ret);
    exit (-1);
  }

  FILE *fp = fopen64 (fs.name.aschar(), "r");
  if (!fp) {
    perror (fs.name.aschar());
    (void)deflateEnd(&strm);
    throw FileReadException();
  }

  do {
    strm.avail_in = fread (in, 1, chunkSize, fp);
    if (ferror(fp)) {
      fprintf (stderr, "Error reading from %s\n", fs.name.aschar());
      fclose (fp);
      (void)deflateEnd(&strm);
      throw FileReadException();
    }
    flush = feof(fp) ? Z_FINISH : Z_NO_FLUSH;
    strm.next_in = in;

    do {
      strm.avail_out = chunkSize;
      strm.next_out = out;
      ret = deflate (&strm, flush);
      assert (ret != Z_STREAM_ERROR);
      compressedSize += chunkSize - strm.avail_out;
    } while (strm.avail_out == 0);
  } while (flush != Z_FINISH);

  (void)deflateEnd(&strm);
  fclose (fp);

  // printf ("%s  size = %Lu  compressed = %Lu  ratio = %f\n",
  //	  fs.name.aschar(), fs.size, compressedSize,
  //	  (double)compressedSize/fs.size);
  fs.size = compressedSize;
}


// Given:  vital stats on a regular file or symbolic link.
// Effect:  add that file to flist.
//          If compressSizes, size is set based on zlib estimate.
// Any other file types cause exit.
static void addFile (const Dstr &fname, const struct stat64 &s,
		     FileList &flist, bool fixDots, bool compressSizes) {
  FileStuff fs;
  fs.name = fname;
  fs.size = s.st_size;
  if (fixDots)
    fixDotDirs (fs.name);

  if (S_ISREG (s.st_mode)) {
    bool retry;
    do {
      retry = false;
      try {
        getHash (MHASH_SHA512, fs.name, fs.hash);
      } catch (FileReadException) {
        Dstr prompt;
        prompt.pruser ("[a]bort, [r]etry, or [i]gnore?", "a");
        switch (prompt[0]) {
        case 'r':
          retry = true;
          break;
        case 'i':
          return;
        default:
          throw;
        }
      }
    } while (retry);

    if (compressSizes)
      compressSize (fs);

  } else if (!S_ISLNK(s.st_mode)) {
    fprintf (stderr, "ERROR: wEiRd file type: %s\n", fs.name.aschar());
    fprintf (stderr, "TA only does regular files, directories, and symbolic links.\n");
    exit (-1);
  }

  flist.push_back (fs);
}


// Given:  a directory name in dirname.
// Effect:  recursively add contents of that directory to flist.
// Directories themselves are not put on the list.
// Directories must have & always get trailing slashes.
static void addDir (const Dstr &dirname, FileList &flist, bool fixDots,
		    bool compressSizes) {
  assert (dirname.back() == '/');
  DIR *dirp = opendir(dirname.aschar());
  if (!dirp) {
    perror (dirname.aschar()); // opendir failed
    exit (-1);
  } else {
    for (dirent64 *dp = readdir64(dirp); dp != NULL;
    dp = readdir64(dirp)) {
      if ((!strcmp (dp->d_name, ".")) ||
          (!strcmp (dp->d_name, "..")))
        continue;
      else {
        struct stat64 s;
        Dstr buf (dirname);
        buf += dp->d_name;
        if (lstat64 (buf.aschar(), &s) == 0) {
          if (S_ISDIR (s.st_mode)) {
            /* Recurse directory */
            buf += '/';
            addDir (buf, flist, fixDots, compressSizes);
          } else
            addFile (buf, s, flist, fixDots, compressSizes);
        } else {
          perror (buf.aschar()); // lstat failed
          exit (-1);
        }
      }
    }
    closedir(dirp);
  }
}


// Given:  a file or directory name in dirname.
// Effect:  add that file or recursively add contents of that directory to
//          flist.
//          . and .. references are fixed only if fixDots is true.
//          If compressSizes, size is set based on zlib estimate.
// Directories themselves are not put on the list.
// Directories need not have trailing slashes.
void addSomething (const Dstr &fname, FileList &flist, bool fixDots,
		   bool compressSizes) {
  struct stat64 s;
  if (lstat64 (fname.aschar(), &s) == 0) {
    if (S_ISDIR (s.st_mode)) {
      Dstr dirname (fname);
      slashterm (dirname);
      addDir (dirname, flist, fixDots, compressSizes);
    } else
      addFile (fname, s, flist, fixDots, compressSizes);
  } else {
    perror (fname.aschar()); // lstat failed
    exit (-1);
  }
}


// Attempt to write hashes to workingdir/ta-hashes.txt.
// In case of failure, hash file is removed and FileWriteException is thrown.
// Written hashes are flist[0] through flist[ptr-1].
void write_hashes (const Dstr &workingdir, const FileList &flist,
FileListIndex ptr) throw (FileWriteException) {
  printf ("Writing hashes\n");
  Dstr hashfname (workingdir);
  hashfname += "ta-hashes.txt";
  // Got inexplicable ESPIPE (Illegal seek) when this file did not exist.
  // rm ("ta-hashes.txt", workingdir);
  unlink (hashfname.aschar());
  try {
    FILE *fp = fopen64 (hashfname.aschar(), "w");
    if (!fp) {
      perror (hashfname.aschar());
      throw FileWriteException();
    } else {
      if (fprintf (fp, "1\n") <= 0) {
	fprintf (stderr, "fprintf returned failure\n");
	fclose (fp);
	throw FileWriteException();
      }
      for (FileListIndex i=0; i<ptr; ++i)
	if (!(flist[i].hash.isNull()))
	  if (fprintf (fp, "%s %s\n", flist[i].hash.aschar(), flist[i].name.aschar()) <= 0) {
            fprintf (stderr, "fprintf returned failure\n");
 	    fclose (fp);
            throw FileWriteException();
          }
      if (fclose (fp)) {
	perror (hashfname.aschar());
	throw FileWriteException();
      }
      // Don't care if chmod fails
      chmod (hashfname.aschar(), 0644);
    }
  } catch (FileWriteException) {
    // rm can fail here with "Read-only file system" (errno = 30) even if
    // the file doesn't exist.
    unlink (hashfname.aschar());
    throw;
  }
}


// Given:  original filename (regular file, directory or symlink) and wdpath.
// Effect:  delete the file from the working directory (NOT THE ORIGINAL).
// If the file does not exist, soldier on.  All other errors result in
// exit.
void rm (const Dstr &filename, const Dstr &wdpath) {
  Dstr target;
  getTarget (filename, wdpath, target);
  printf ("Remove %s\n", target.aschar());
  if (remove (target.aschar()) != 0) {
    // Sometimes perror itself causes an ESPIPE (Illegal seek), so preserve
    // the original errno for checking below.
    const int errno_copy (errno);
    perror (target.aschar());
    if (errno_copy != ENOENT) {
      printf ("errno = %d\n", errno_copy);
      exit (-1);
    }
  }
}
