// $Id: ta.cc 4847 2013-02-08 00:54:40Z flaterco $

/*
    TA (Transparent Archivist) (C) 2005 David Flater.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mount.h>
#include <math.h>
#include "ta-util.hh"
#include <unistd.h>
#include <locale.h>
#include <langinfo.h>
#include <set>


// Choose your file system.
enum FileSystemEnum {iso, isoj, isorr, ext2, le2a, le2f, udf, squashfs};


// Test for read-only file system (those that use a mk...fs program and need
// a temporary copy of everything).
bool ReadOnlyFSp (FileSystemEnum  e) {
  return e == iso || e == isoj || e == isorr || e == squashfs;
}


// Force permissions on all directories.
void force_dir_perms (const Dstr &dirname) {
  // This syntax works for GNU find version 4.1.7.
  char *args[] = {"find", dirname.aschar(), "-mindepth", "1",
	          "-type", "d", "-exec", "chmod", "755", "{}", ";", NULL};
  sys (args);
}


// Given:  filename.
// Effect:  Copy into working dir at wdpath, creating any missing
//   parent directories.  If filename specifies a directory, nothing
//   happens.
//
// If keep_props is true, try to preserve ownership, permissions, and
// timestamps.  If keep_props is false, only try to preserve
// timestamps.
//
// If the cp command returns a nonzero exit status, this is thrown as
// an exception of type int.
//
// When run as root, cp tries to preserve ownership and returns
// nonzero if it fails (e.g. a vfat file system doesn't allow it).
// When run as non-root, cp does not even try to preserve ownership--
// it just returns success!

void cp (const Dstr &filename, const Dstr &wdpath, bool keep_props,
bool force_perms) throw (int) {
  assert (filename.length());
  if (filename.back() != '/') {
    printf ("Copy %s into %s\n", filename.aschar(), wdpath.aschar());

    // Copying a file is more complicated than you would think.  GNU
    // coreutils 5.92 devote 2000 lines of code to it (counting only
    // copy.c and copy.h).

    // This syntax works for GNU cp (coreutils) 5.2.1.

    // Issue:  Refraining from --preserve=mode does NOT get you
    // default permissions.  cp will tighten permissions if umask is
    // less friendly, but it will not relax permissions if umask is
    // fiendlier.  Hence the accursed "permission denied" problem.

    // One option would be to force directory permissions after every
    // copy.  Not very efficient.

    // Another option would be to force directory permissions only
    // when cp fails.  Unfortunately, there is no way to distinguish
    // permission denied from file system full.

    // Best option is just to run as root.

    if (keep_props) {
      if (force_perms) {
        char *args1[] = {"cp", "--parents", "--no-dereference",
                         "--preserve=ownership,timestamps",
                         filename.aschar(), wdpath.aschar(), NULL};
        sys (args1);
        Dstr target;
        getTarget (filename, wdpath, target);
        char *args2[] = {"setperm", target.aschar(), NULL};
        sys (args2);
      } else {
        char *args[] = {"cp", "--parents", "--no-dereference",
			"--preserve=mode,ownership,timestamps",
                        filename.aschar(), wdpath.aschar(), NULL};
        sys (args);
      }
    } else {
      char *args[] = {"cp", "--parents", "--no-dereference",
                      "--preserve=timestamps",
                      filename.aschar(), wdpath.aschar(), NULL};
      sys (args);
    }
  }
}


// Ditto, but with multiple source files to speed up the copy.
// Use of -p option still makes it slow.
void cpLots (const FileList &flist, const FileListIndex oldEnd,
	     const FileListIndex newEnd, const Dstr &wdpath, bool keep_props,
	     bool force_perms) throw (int) {

  // The tricky bit here is not to exceed the limits on argument list length
  // (very likely) or number of arguments (?).  Various overheads eat into
  // ARG_MAX, so just leave a big safety margin.
  static long commandLineLengthLimit = 0;
  if (!commandLineLengthLimit) {
    if ((commandLineLengthLimit = sysconf(_SC_ARG_MAX)) < 0) {
      printf ("sysconf(_SC_ARG_MAX) failed; making a wild guess.\n");
      commandLineLengthLimit = 4096;
    }
    printf ("ARG_MAX = %ld\n", commandLineLengthLimit);
    commandLineLengthLimit /= 2;
  }

  FileListIndex i = oldEnd;
  while (i < newEnd) {
    long commandLineLength = 0;
    std::vector<char*> args;
    args.push_back ("cp");
    args.push_back ("--parents");
    args.push_back ("--no-dereference");
    if (keep_props) {
      if (force_perms)
	args.push_back ("--preserve=ownership,timestamps");
      else
	args.push_back ("--preserve=mode,ownership,timestamps");
    } else
      args.push_back ("--preserve=timestamps");
    while ((i < newEnd) && (commandLineLength < commandLineLengthLimit)) {
      const Dstr &filename (flist[i++].name);
      assert (filename.length());
      if (filename.back() != '/') {
	args.push_back (filename.aschar());
	commandLineLength += filename.length() + 1;
      }
    }
    args.push_back (wdpath.aschar());
    args.push_back (NULL);

    if (commandLineLength) { // No go if all source files were directories.
      printf ("Copying files into %s\n", wdpath.aschar());
      sys (&(args[0]));
    }
  }

  // -p will still be slow.
  if (keep_props && force_perms && (oldEnd < newEnd)) {
    printf ("Fixing permissions...\n");
    for (i=oldEnd; i<newEnd; ++i) {
      const Dstr &filename (flist[i].name);
      if (filename.back() != '/') {
        Dstr target;
        getTarget (filename, wdpath, target);
        char *args2[] = {"setperm", target.aschar(), NULL};
        sys (args2);
      }
    }
  }
}


// Purge empty directories below wdpath.
void purgeEmptyDirs (const Dstr &wdpath) {
  printf ("Purging empty directories\n");
  // This syntax works for GNU find version 4.1.7.
  char *args[] = {"find", wdpath.aschar(), "-depth", "-mindepth", "1", 
    "-empty", "-type", "d", "-exec", "rmdir", "{}", ";", NULL};
  sys (args);
}


// Given:  original filename and wdpath.
// Effect:  rm that file from the working directory, then clean empty
// directories from wdpath.  If the file does not exist, just clean
// empty directories.
void rmclean (const Dstr &filename, const Dstr &wdpath) {
  rm (filename, wdpath);
  purgeEmptyDirs (wdpath);
}


// rm -rf a directory.
void nukedir (const Dstr &dirname) {
  // 20060701
  // rm -rf CAN fail if a directory is not writable.
  // (Never fails if run by root.)
  // chmod fails if the target does not exist.
  // Anyway, this function doesn't fail except when rm fails.
  try {
    char *args1[] = {"chmod", "-Rf", "u+rwx", dirname.aschar(), NULL};
    sys (args1);
  } catch (int retstatus) {
    ;
  }
  char *args2[] = {"rm", "-rf", dirname.aschar(), NULL};
  sys (args2);
}


// Given:  a directory tree containing files 0 through oldEnd-1
// Effect:  modify that directory tree to contain files 0 through newEnd-1
// You can create a new tree by specifying oldEnd = 0.
//
// See cp regarding keep_props.

void adjustWorkingDir (const Dstr &wdpath, const FileList &flist,
                       FileListIndex oldEnd, FileListIndex newEnd,
                       bool keep_props, bool force_perms) {
  assert (oldEnd <= flist.size());
  assert (newEnd <= flist.size());
  if (oldEnd > newEnd) {
    while (oldEnd > newEnd)
      rm (flist[--oldEnd].name, wdpath);
    purgeEmptyDirs (wdpath);

    // It is not sufficient to put the directories in flist and remove
    // them in reverse order.  For example, if you were archiving the
    // directory /home/somebody/foo and then backed up, you would
    // remove /wdpath/home/somebody/foo, but /wdpath/home/somebody and
    // /wdpath/home would still be there.

  } else
    cpLots (flist, oldEnd, newEnd, wdpath, keep_props, force_perms);
}


// Delete transferred files from flist.
void purge (FileList &flist, FileListIndex ptr) {
  assert (ptr <= flist.size());
  if (ptr == flist.size())
    flist.clear();
  else
    flist.erase (flist.begin(), flist.begin()+ptr);
}


// unlink with error check.  Quite often, if this fails, it will only be
// adding one more frantic error message to an already dire situation (i.e.,
// the next statement was going to be an exit -1 anyway).
void checkedUnlink (const Dstr &path) {
  if (unlink (path.aschar()) != 0) {
    perror (path.aschar());
    fprintf (stderr, "Unexpected error deleting %s... exiting.\n", path.aschar());
    exit (-1);
  }
}


// Check whether we are in a no-win for fitting stuff on the media.
// ptr is the index of the first file that won't fit (= the number of
//   files that *will* fit).
// minAcceptableFit is the minimum number of files that *must* fit.
// If we are in trouble, print errors and return true.
// Otherwise print nothing and return false.
bool nothingFits (FileList &flist,
		  FileListIndex minAcceptableFit,
		  FileListIndex ptr) {
  assert (minAcceptableFit >= 1);
  if (ptr < minAcceptableFit) {
    if (ptr < minAcceptableFit-1)
      fprintf (stderr, "ERROR: -r files don't fit on media.\n");
    else
      fprintf (stderr, "ERROR: can't fit %s on media.\n",
	flist[minAcceptableFit-1].name.aschar());
    return true;
  }
  return false;
}


// Clean up mess from buildWritableImage.  One or more of the parameters can
// be null in order to skip steps if we had a failure partway through.
void cleanupWritable (const Dstr &loopdev,
		      const Dstr &mountdev,
		      const Dstr &mountpoint) {
  if (!mountpoint.isNull())
    if (umount (mountpoint.aschar()) != 0) {
      perror (mountpoint.aschar());
      fprintf (stderr, "ERROR: can't unmount %s\n", mountpoint.aschar());
      // We are in trouble now
      exit (-1);
    }

  // Tear down crypto
  if (!mountdev.isNull())
    if (loopdev != mountdev) {
      char *args[] = {"cryptsetup", "luksClose", "TA", NULL};
      sys (args);
    }

  // Tear down loopback
  if (!loopdev.isNull()) {
    char *args[] = {"losetup", "--detach", loopdev.aschar(), NULL};
    sys (args);
  }
}


// Convenience function used by buildWritableImage.
// ptr is the index of the first file that won't fit.
void checkWritableFit (const Dstr &loopdev,
                       const Dstr &mountdev,
		       const Dstr &workingdir,
                       const Dstr &imgfilename,
		       FileList &flist,
		       FileListIndex minAcceptableFit,
		       FileListIndex ptr) {
  if (nothingFits (flist, minAcceptableFit, ptr)) {
    cleanupWritable (loopdev, mountdev, workingdir);
    checkedUnlink (imgfilename.aschar());
    exit (-1);
  }
}


// Set up a loop on an image file.  The name of the loop device used is
// returned in loopdev_out.
void loopimg (const Dstr &imgfilename, Dstr &loopdev_out) {
  Dstr loseoutfname ("/tmp/ta_deleteme_pid");
  loseoutfname += getpid();
  loseoutfname += ".txt";
  char *args[] = {"losetup", "--find", "--show", imgfilename.aschar(), NULL};
  sys (loseoutfname.aschar(), args);
  FILE *fp = fopen (loseoutfname.aschar(), "r");
  if (!fp) {
    perror (loseoutfname.aschar());
    checkedUnlink (imgfilename.aschar());
    // Possibly leaking a loopback device
    exit (-1);
  }
  loopdev_out.getline (fp);
  fclose (fp);
  checkedUnlink (loseoutfname.aschar());
  if (loopdev_out.isNull()) {
    fprintf (stderr, "ERROR: losetup failed to produce output\n");
    checkedUnlink (imgfilename.aschar());
    // Possibly leaking a loopback device
    exit (-1);
  }
  printf ("Using %s\n", loopdev_out.aschar());
}


// Make ext2/le2a/le2f on the image file that is already created, mount it
// and remove lost+found.  The name of the loop device used is returned in
// loopdev_out.  The name of the device actually mounted is returned in
// mountdev_out.  Sometimes these are the same.
void mke2 (const Dstr &imgfilename, const Dstr &mountpoint,
	   Dstr &loopdev_out, Dstr &mountdev_out,
           FileSystemEnum fstype, bool gurue2) {

  loopimg (imgfilename, loopdev_out);

  // Configure crypto
  switch (fstype) {
  case ext2:
    mountdev_out = loopdev_out;
    break;
  case le2a:
  case le2f:
    {
      // Const issues
      Dstr cipher (fstype == le2a ? "aes-xts-plain" : "twofish-xts-plain");
      // XTS uses two keys of the same size, so -s 512 is really 256-bit crypto.
      char *args[] = {"cryptsetup", "-c", cipher.aschar(), "-s", "512", "-q",
		      "luksFormat", loopdev_out.aschar(), NULL};
      sys (args);
    }
    printf ("LUKS formatting complete.  Now to open the volume we need the password again.\n");
    {
      char *args[] = {"cryptsetup", "luksOpen", loopdev_out.aschar(), "TA", NULL};
      sys (args);
    }
    mountdev_out = "/dev/mapper/TA";
    break;
  default:
    assert (false);
  }

  // mke2fs
  if (gurue2) {
    char *args[] = {"mke2fs", "-m", "0", "-N", "1", "-O",
                    "none,sparse_super,filetype", "-I", "128",
		    mountdev_out.aschar(), NULL};
    sys (args);
  } else {
    char *args[] = {"mke2fs", "-m", "0", mountdev_out.aschar(), NULL};
    sys (args);
  }

  // tune2fs
  {
    char *args[] = {"tune2fs", "-c", "0", "-i", "0", mountdev_out.aschar(),
		    NULL};
    sys (args);
  }

  if (mount (mountdev_out.aschar(), mountpoint.aschar(), "ext2", 0, NULL) != 0) {
    // Most likely mode of failure is problem with the mount point.
    perror (mountpoint.aschar());
    fprintf (stderr, "ERROR: can't mount %s on %s\n", mountdev_out.aschar(), mountpoint.aschar());
    cleanupWritable (loopdev_out, mountdev_out, (char *)NULL);
    checkedUnlink (imgfilename.aschar());
    exit (-1);
  }

  rm ("lost+found", mountpoint);
}


// Effect:  image created; saved files are removed from flist.
void buildWritableImage (const Dstr &imgfilename, const Dstr &workingdir,
			 FileList &flist, FileList &rlist, ta_t discsize,
                         FileSystemEnum fstype, bool force_perms, bool gurue2,
                         bool nornd, const Dstr &codeset) {

  if (!discsize) {
    fprintf (stderr, "ERROR: discsize is zero!\n");
    exit (-1);
  }
  assert (!(flist.empty()));

  FileListIndex minAcceptableFit (rlist.size() + 1);
  flist.insert (flist.begin(), rlist.begin(), rlist.end());

  createFile (imgfilename, discsize,
    !nornd && (fstype == le2a || fstype == le2f));

  Dstr loopdev, mountdev;

  switch (fstype) {
  case ext2:
  case le2a:
  case le2f:
    mke2 (imgfilename, workingdir, loopdev, mountdev, fstype, gurue2);
    break;

  case udf:
    // Must use the source (sigh).  media-type is a misleading code from
    // which mkudffs infers the settings that we supposedly want.
    // 20080125: Was using dvd successfully, now this:
    //   UDF-fs: Partition marked readonly; forcing readonly mount
    //   (even if rw is specified at mount time)
    {
      char *args[] = {"mkudffs", "--media-type=hd", imgfilename.aschar(),
                      NULL};
      sys (args);
    }
    loopimg (imgfilename, loopdev);
    mountdev = loopdev;
    {
      Dstr opts ("iocharset=");
      opts += codeset;
      if (mount (mountdev.aschar(), workingdir.aschar(), "udf", 0, opts.aschar()) != 0) {
        // Most likely mode of failure is problem with the mount point.
        perror (workingdir.aschar());
        fprintf (stderr, "ERROR: can't mount %s on %s\n", mountdev.aschar(), workingdir.aschar());
        cleanupWritable (loopdev, mountdev, (char *)NULL);
        checkedUnlink (imgfilename.aschar());
        exit (-1);
      }
    }
    rm ("lost+found", workingdir);
    break;

  default:
    assert (0);
  }

  // Fill up /mnt
  FileListIndex ptr = 0;
  {
    try {
      while (ptr < flist.size()) {
	cp (flist[ptr].name, workingdir, true, force_perms);
	++ptr;
      }
    } catch (int retstatus) {
      rmclean (flist[ptr].name, workingdir);
    }
  }
  checkWritableFit (loopdev, mountdev, workingdir, imgfilename, flist, minAcceptableFit, ptr);

  // Attempt to write hashes.  If it fills up again, delete another file
  // and repeat.
  bool diskfull;
  do {
    diskfull = false;
    try {
      write_hashes (workingdir, flist, ptr);
    } catch (FileWriteException) {
      diskfull = true;
      printf ("Disk too full to write hashes\n");
      checkWritableFit (loopdev, mountdev, workingdir, imgfilename, flist, minAcceptableFit, --ptr);
      rmclean (flist[ptr].name, workingdir);
    }
  } while (diskfull);

  if (force_perms)
    force_dir_perms (workingdir);

  cleanupWritable (loopdev, mountdev, workingdir);
  purge (flist, ptr);
}


// Algorithm 1
// Given:  flist, ratio and discsize.
// Return:  pointer to the first file that won't fit (or size() if all fit)
//          based on assumption that all files compress at the specified
//          ratio.
// Guarantees only ptr >= 0 and ptr <= flist.size().
// Alg 1 is unstable when compression ratios vary drastically.
FileListIndex fit_ratio (const FileList &flist,
			 double ratio,
			 ta_t discsize) {
  ta_t before_size = 0;
  FileListIndex ptr;
  for (ptr = 0; ptr < flist.size(); ++ptr) {
    before_size += flist[ptr].size;
    if (before_size * ratio > discsize)
      break;
  }
  return ptr;
}


// Algorithm 0
// Given:  willFit, wontFit, flist, discsize
// Return:  bisection try for newEnd.
//          If the right side of the bracket is not yet established,
//          limit size of try based on assumed maximum 90% compression.
// Guarantees return > willFit (or > 0) and return < wontFit (or <= size).
FileListIndex bisect (bool have_willFit, FileListIndex willFit,
		      bool have_wontFit, FileListIndex wontFit,
		      const FileList &flist,
                      ta_t discsize) {
  FileListIndex ptr = 0;
  if (have_willFit)
    ptr = willFit;
  if (have_wontFit)
    ptr += wontFit;
  else
    ptr += flist.size() + 1;
  ptr /= 2;

  // Limiter
  if (!have_wontFit) {
    FileListIndex maxPtr = fit_ratio (flist, 0.1, discsize);
    // If we have already busted that limit, shrug and proceed.
    if ((maxPtr < ptr) && ((!have_willFit) || (have_willFit && (maxPtr > willFit)))) {
      printf ("grepline:  bisect size limiter reducing %lu to %lu\n",
	      ptr, maxPtr);
      ptr = maxPtr;
    }
  }

  assert (ptr > 0);
  if (have_willFit)
    assert (ptr > willFit);
  assert (ptr <= flist.size());
  if (have_wontFit)
    assert (ptr < wontFit);
  return ptr;
}


// Algorithm 2
// Given:  left and right bounds with known before and after sizes
// Return:  pointer to the first file that won't fit based on linear
//          interpolation between the two bounds, or the neighboring
//          file if that one is a bound.
// Guarantees return > willFit and return < wontFit.
// Like alg 1, alg 2 employs a bogus same-ratio assumption, but it
// applies only to the files within the bracket.  Alg 2 is stable but
// suboptimal when compression ratios vary drastically.
FileListIndex fit_linear (const FileList &flist,
			  FileListIndex willFit,
			  ta_t left_before_size,
			  ta_t left_after_size,
			  FileListIndex wontFit,
			  ta_t right_before_size,
			  ta_t right_after_size,
			  ta_t discsize) {
  assert (willFit + 1 < wontFit);
  assert (right_before_size >= left_before_size);

  // While it is possible that a file system could actually get *smaller*
  // when more data are added, the selection of our bracket by
  // buildReadOnlyImage prevents an inversion from ever appearing here.
  assert (left_after_size <= discsize);
  assert (right_after_size > discsize);

  FileListIndex ptr = willFit;
  if (right_before_size == left_before_size)
    ptr = (willFit + wontFit) / 2;
  else {
    double ratio = (double)(right_after_size - left_after_size) /
                           (right_before_size - left_before_size);
    ta_t bracket_size = flist[ptr].size;
    bracket_size += flist[++ptr].size;
    while ((left_after_size + bracket_size * ratio <= discsize) &&
	   (ptr + 1 < wontFit))
      bracket_size += flist[++ptr].size;
  }
  assert (ptr > willFit && ptr < wontFit);
  return ptr;
}


// Sum the file sizes up to ptr-1.
ta_t getFlistSize (const FileList &flist, FileListIndex ptr) {
  ta_t before_size = 0;
  for (FileListIndex i=0; i<ptr; ++i)
    before_size += flist[i].size;
  return before_size;
}


// Verify that workingdir is empty, halt if not.
void WorkingDirSafetyCheck (const Dstr &workingdir) {
  assert (workingdir.back() == '/');
  DIR *dirp = opendir(workingdir.aschar());
  if (!dirp) {
    perror (workingdir.aschar()); // opendir failed
    exit (-1);
  } else {
    for (dirent64 *dp = readdir64(dirp); dp != NULL;
    dp = readdir64(dirp)) {
      if ((!strcmp (dp->d_name, ".")) ||
          (!strcmp (dp->d_name, "..")))
        continue;
      else {
        fprintf (stderr, "ERROR: workingdir is not empty!\n");
        exit (-1);
      }
    }
    closedir(dirp);
  }
}


// Effect:  image created; saved files are removed from flist.
// Complication:  mkisofs can predict size without doing all the work;
//                mksquashfs cannot.
void buildReadOnlyImage (const Dstr &imgfilename, const Dstr &workingdir,
			 FileList &flist, FileList &rlist, ta_t discsize,
			 double &ratio, FileSystemEnum fstype,
			 bool force_perms, const Dstr &codeset) {

  Dstr sizefname ("/tmp/ta_deleteme_pid");
  sizefname += getpid();
  sizefname += ".txt";

  if (!discsize) {
    fprintf (stderr, "ERROR: discsize is zero!\n");
    exit (-1);
  }
  assert (!(flist.empty()));

  FileListIndex minAcceptableFit (rlist.size() + 1);
  flist.insert (flist.begin(), rlist.begin(), rlist.end());

  Dstr before (workingdir);
  before += "before/";
  if (mkdir (before.aschar(), 0755)) {
    perror (before.aschar());
    exit (-1);
  }

  Dstr tempimgfilename (imgfilename);
  tempimgfilename += "temp";

  bool have_willFit = false, have_wontFit = false, past_limit = false;
  FileListIndex willFit = 0, wontFit = 0, oldEnd = 0;
  ta_t left_before_size = 0, left_after_size = 0,
       right_before_size = 0, right_after_size = 0;
  unsigned long iterations = 0;
  unsigned stupidity = 0;
  const FileListIndex greedyAlgorithmIterationsLimit = log2(flist.size()) + 1;
  printf ("Have %lu files\n", (unsigned long)(flist.size()));
  printf ("Assigning greedy algorithm iterations limit = %lu\n",
	   greedyAlgorithmIterationsLimit);

  while (!((have_willFit && have_wontFit && (willFit + 1 == wontFit)) ||
           (have_willFit && (willFit == flist.size())))) {
    if (have_wontFit)
      assert (wontFit > 1);
    if (have_willFit)
      assert (willFit > 0);
    if (have_willFit && have_wontFit)
      assert (willFit + 1 < wontFit);

    // I want this output in case there's an assertion failure before we get
    // to the 'grepline'.
    if (have_willFit)
      printf ("Left = %lu  ", willFit);
    else
      printf ("Left is blank  ");
    if (have_wontFit)
      printf ("Right = %lu\n", wontFit);
    else
      printf ("Right is blank\n");

    if (!past_limit && iterations >= greedyAlgorithmIterationsLimit) {
      printf ("Have reached greedy algorithm iterations limit.\n"
              "Switching to binary search.\n");
      past_limit = true;
    }

    int algorithm_used = -1;
    FileListIndex newEnd;
    if (past_limit) {
      algorithm_used = 0;
      newEnd = bisect (have_willFit, willFit, have_wontFit, wontFit,
		       flist, discsize);
    } else if (have_willFit && have_wontFit) {
      algorithm_used = 2;
      newEnd = fit_linear (flist, willFit, left_before_size, left_after_size,
			   wontFit, right_before_size, right_after_size,
			   discsize);
      // Alg 2 is stable but could still suffer from slow convergence.
      // For now, track and report stupidity but don't intervene.
      if (newEnd == willFit + 1 || newEnd == wontFit - 1)
        ++stupidity;
      else
        stupidity = 0;
    } else {
      algorithm_used = 1;
      newEnd = fit_ratio (flist, ratio, discsize);
      if (have_willFit && newEnd <= willFit) {
        ++stupidity;
        newEnd = willFit + 1;
      } else if (newEnd == 0) {
        ++stupidity;
        newEnd = 1;
      } else if (have_wontFit && newEnd >= wontFit) {
        ++stupidity;
        newEnd = wontFit - 1;
      } else
        stupidity = 0;
      if (stupidity >= 3) {
        algorithm_used = 10;
        newEnd = bisect (have_willFit, willFit, have_wontFit, wontFit,
			 flist, discsize);
      }
    }

    assert (newEnd > 0);
    if (have_willFit)
      assert (newEnd > willFit);
    assert (newEnd <= flist.size());
    if (have_wontFit)
      assert (newEnd < wontFit);

    printf ("grepline:  ");
    if (have_willFit)
      printf ("Left = %lu", willFit);
    else
      printf ("Left is blank");
    if (have_wontFit)
      printf ("  Right = %lu", wontFit);
    else
      printf ("  Right is blank");
    printf ("  Size = %lu  Iterations = %lu  Limit = %lu  Stupidity = %u  Algorithm = %d  Trying = %lu\n",
      (unsigned long)(flist.size()), iterations,
      greedyAlgorithmIterationsLimit, stupidity, algorithm_used, newEnd);

    ta_t before_size, after_size;
    before_size = getFlistSize (flist, newEnd);
    adjustWorkingDir (before, flist, oldEnd, newEnd, (fstype != iso), force_perms);
    write_hashes (before, flist, newEnd);
    if (force_perms)
      force_dir_perms (before);

    fflush (stdout);
    fflush (stderr);

    // DO NOT CHANGE THE ISO-LEVEL FOR ISOJ!
    // WINDOWS XP IGNORES JOLIET IF THE DISC IS ISO-9660:1999!

    // This syntax works for mkisofs 2.01.
    switch (fstype) {
    case isoj:
      {
	char *args[] = {"mkisofs", "-print-size", "-no-pad", "-iso-level",
                        "1", "-D", "-jcharset", codeset.aschar(),
                        "-joliet-long", before.aschar(), NULL};
	sys (sizefname.aschar(), args);
      }
      break;
    case iso:
      {
	char *args[] = {"mkisofs", "-print-size", "-no-pad",
			"-iso-level", "4", before.aschar(), NULL};
	sys (sizefname.aschar(), args);
      }
      break;
    case isorr:
      {
	char *args[] = {"mkisofs", "-print-size", "-no-pad",
			"-iso-level", "4", "-R", before.aschar(), NULL};
	sys (sizefname.aschar(), args);
      }
      break;
    case squashfs:
      {
        // Don't add -nopad here; it breaks loopback.
        char *args[] = {"mksquashfs", before.aschar(),
			tempimgfilename.aschar(), "-no-progress",
			NULL};
        try {
  	  sys (args);
        } catch (int retstatus) {
          fflush (stdout);
          fflush (stderr);
          fprintf (stderr, "ta: mksquashfs exited with status %d\n", retstatus);
          fflush (stderr);
          throw;
        }
      }
      break;
    default:
      assert (0);
    }

    fflush (stdout);
    fflush (stderr);

    // For ISO, get predicted size; for squashfs, get actual size.
    if (fstype == squashfs) {
      struct stat64 s;
      if (lstat64 (tempimgfilename.aschar(), &s) == 0)
        after_size = s.st_size;
      else {
	perror (tempimgfilename.aschar()); // lstat failed
        fprintf (stderr, "Unexpected error with squashfs output file... exiting.\n");
	exit (-1);
      }
    } else {
      FILE *fp = fopen (sizefname.aschar(), "r");
      if (!fp) {
        perror (sizefname.aschar());
        fprintf (stderr, "Unexpected error with mkisofs output file... exiting.\n");
        exit (-1);
      }
      assert (fscanf (fp, "%Lu", &after_size) == 1);
      fclose (fp);
      after_size *= 2048; // mkisofs reports sectors not bytes
    }

    printf ("discsize = %Lu   after_size = %Lu   ", discsize, after_size);
    if (after_size <= discsize) {
      printf ("It'll fit.\n");
      have_willFit = true;
      willFit = newEnd;
      left_before_size = before_size;
      left_after_size = after_size;
      if (fstype == squashfs) {
        if (rename (tempimgfilename.aschar(), imgfilename.aschar()) != 0) {
          perror ("rename");
          fprintf (stderr, "Unexpected error renaming %s to %s... exiting.\n",
		   tempimgfilename.aschar(), imgfilename.aschar());
          exit (-1);
        }
      }
    } else {
      printf ("It won't fit.\n");
      have_wontFit = true;
      wontFit = newEnd;
      right_before_size = before_size;
      right_after_size = after_size;
      if (fstype == squashfs)
	checkedUnlink (tempimgfilename.aschar());
      if (nothingFits (flist, minAcceptableFit, wontFit-1)) {
        nukedir (before);
        if (fstype != squashfs)
          checkedUnlink (sizefname.aschar());
	exit (-1);
      }
    }

    oldEnd = newEnd;

    // Ratio can go to infinity if we have a bunch of zero-length files.
    // Don't let that happen.
    if (before_size)
      ratio = (double)after_size / before_size;
    else
      ratio = 1.0;
    printf ("New ratio is %f\n", ratio);
    ++iterations;
  }

  printf ("Converged:  %lu iterations, %lu files.\n", iterations, willFit);

  // If squashfs then we're done already.
  if (fstype != squashfs) {
    checkedUnlink (sizefname.aschar());
    printf ("Now making %s for real.\n", imgfilename.aschar());

    adjustWorkingDir (before, flist, oldEnd, willFit, (fstype != iso), force_perms);
    write_hashes (before, flist, willFit);
    if (force_perms)
      force_dir_perms (before);

    // This syntax works for mkisofs 2.01.
    switch (fstype) {
    case isoj:
      {
	char *args[] = {"mkisofs", "-o", imgfilename.aschar(), "-no-pad",
			"-iso-level", "1", "-D", "-jcharset", codeset.aschar(),
			"-joliet-long", before.aschar(), NULL};
	sys (args);
      }
      break;
    case iso:
      {
	char *args[] = {"mkisofs", "-o", imgfilename.aschar(), "-no-pad",
			"-iso-level", "4", before.aschar(), NULL};
	sys (args);
      }
      break;
    case isorr:
      {
	char *args[] = {"mkisofs", "-o", imgfilename.aschar(), "-no-pad",
			"-iso-level", "4", "-R", before.aschar(), NULL};
	sys (args);
      }
      break;
    default:
      assert (0);
    }
  }
  nukedir (before);
  purge (flist, willFit);
}


// Does not return.
void usagebarf () {
  fprintf (stderr, "\
Usage: ta [options] discsize filesystem workingdir imagedir src [src...]\n\
\n\
Options:\n\
-l      Tweak ext2/le2a/le2f to maximize usable space.\n\
-nornd  Don't initialize encrypted volumes with random data.\n\
-p      Force file permissions to reasonable defaults.\n\
-r src  Replicate file src in *every* image.\n\
-w      Wait for confirmation after completing each image.\n\
\n\
discsize:  cd74, cd80, dvd+r, dvd+rdl, or an arbitrary size specified in bytes.\n\
filesystem:  iso, isoj, isorr, ext2, le2a, le2f, squashfs or udf.\n\
workingdir:  for ext2, le2a, le2f or udf this is just a mount point that we can\n\
  use.  For others, this must be an existing, empty directory that we can fill\n\
  up and then wipe clean.\n\
imagedir:  disc images will be written here, overwriting any files that are in\n\
  the way.  Make sure it is on a file system that can handle big files if you\n\
  are creating DVD-sized images.\n\
src:  stuff to archive.  Should usually be a directory, but you can do single\n\
  files if you want.\n\
");
  exit (-1);
}


void wait_for_user () {
  fprintf (stderr, "\nPress Enter to continue....\n");
  Dstr junk;
  junk.getline (stdin);
}


void fixDuplicates (FileList &flist, FileList &rlist) {
  std::set<Dstr> fset, rset;
  for (FileListIndex i=0; i<rlist.size(); ) {
    if (rset.count(rlist[i].name)) {
      fprintf (stderr, "Warning:  Removing duplicate %s from -r list\n", rlist[i].name.aschar());
      rlist.erase (rlist.begin()+i, rlist.begin()+i+1);
    } else {
      rset.insert (rlist[i++].name);
    }
  }
  for (FileListIndex i=0; i<flist.size(); ) {
    if (rset.count(flist[i].name)) {
      fprintf (stderr, "Warning:  %s is in -r list; removing it from src list\n", flist[i].name.aschar());
      flist.erase (flist.begin()+i, flist.begin()+i+1);
    } else if (fset.count(flist[i].name)) {
      fprintf (stderr, "Warning:  Removing duplicate %s from src list\n", flist[i].name.aschar());
      flist.erase (flist.begin()+i, flist.begin()+i+1);
    } else {
      fset.insert (flist[i++].name);
    }
  }
}


void GetCodeset (Dstr &codeset, FileSystemEnum fstype) {
  if (fstype == isoj || fstype == udf) {

    // Method suggested by man utf-8 to obtain ambient codeset.
    setlocale (LC_CTYPE, "");
    codeset = nl_langinfo (CODESET);

    // Undo damage so that libc functions won't change behavior in
    // mysterious ways.
    setlocale (LC_CTYPE, "C");

    printf ("Your locale's codeset is %s.\n", codeset.aschar());

    // Translate to syntax understood by mkisofs and mount.
    if (codeset %= "ISO-8859-")
      codeset.repstr ("ISO-", "iso");
    else if (codeset == "UTF-8") {
      if (fstype == isoj) {
        fprintf (stderr, "ERROR:  UTF-8 not supported by isoj.\n");
        exit (-1);
      }
      codeset = "utf8";
    } else {
      fprintf (stderr, "ERROR:  codeset %s not handled.  Cannot do isoj or udf without a codeset mapping.\n", codeset.aschar());
      exit (-1);
    }

    printf ("Name to give to %s is %s.\n",
	    fstype == isoj ? "mkisofs" : "mount", codeset.aschar());
  }
}


int main (int argc, char **argv) {
  printf ("\
    TA (Transparent Archivist) Version %s (C) 2005 David Flater.\n\
\n\
    This program is free software: you can redistribute it and/or modify\n\
    it under the terms of the GNU General Public License as published by\n\
    the Free Software Foundation, either version 3 of the License, or\n\
    (at your option) any later version.\n\
\n\
    This program is distributed in the hope that it will be useful,\n\
    but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
    GNU General Public License for more details.\n\
\n\
    You should have received a copy of the GNU General Public License\n\
    along with this program.  If not, see <http://www.gnu.org/licenses/>.\n\
\n", ta_version);

  if (argc < 6)
    usagebarf();

  int argi = 1;
  bool force_perms = false, pausewait = false, gurue2 = false, nornd = false;
  std::vector<Dstr> replicants;

  for (bool doneOpts (false); !doneOpts; ) {
    if (!strcmp (argv[argi], "-p")) {
      force_perms = true;
      ++argi;
    } else if (!strcmp (argv[argi], "-l")) {
      gurue2 = true;
      ++argi;
    } else if (!strcmp (argv[argi], "-nornd")) {
      nornd = true;
      ++argi;
    } else if (!strcmp (argv[argi], "-w")) {
      pausewait = true;
      ++argi;
    } else if (!strcmp (argv[argi], "-r")) {
      ++argi;
      if (argi >= argc)
        usagebarf();
      replicants.push_back (argv[argi++]);
    } else
      doneOpts = true;
  }
  if (argi + 4 >= argc)
    usagebarf();

  ta_t discsize;
  if (!strcmp (argv[argi], "cd74"))
    discsize = CD74size;
  else if (!strcmp (argv[argi], "cd80"))
    discsize = CD80size;
  else if (!strcmp (argv[argi], "dvd+r"))
    discsize = DVD5size;
  else if (!strcmp (argv[argi], "dvd+rdl"))
    discsize = DVD9size;
  else if (sscanf (argv[argi], "%llu", &discsize) != 1) {
    fprintf (stderr, "Bad discsize specification: %s\n", argv[argi]);
    exit (-1);
  }
  ++argi;

  FileSystemEnum fstype;
  if (!strcmp (argv[argi], "ext2"))
    fstype = ext2;
  else if (!strcmp (argv[argi], "le2a"))
    fstype = le2a;
  else if (!strcmp (argv[argi], "le2f"))
    fstype = le2f;
  else if (!strcmp (argv[argi], "iso"))
    fstype = iso;
  else if (!strcmp (argv[argi], "isoj"))
    fstype = isoj;
  else if (!strcmp (argv[argi], "isorr"))
    fstype = isorr;
  else if (!strcmp (argv[argi], "squashfs"))
    fstype = squashfs;
  else if (!strcmp (argv[argi], "udf")) {
    fprintf (stderr, "WARNING!  UDF support is experimental.  Linux kernel stability problems\nmay occur!\n");
    fstype = udf;
  } else {
    fprintf (stderr, "Unrecognized filesystem: %s\n", argv[argi]);
    fprintf (stderr, "Choices are iso, isoj, isorr, ext2, le2a, le2f, squashfs and udf.\n");
    exit (-1);
  }

  Dstr workingdir (argv[++argi]);
  Dstr imagedir (argv[++argi]);
  slashterm (workingdir);
  slashterm (imagedir);

  if (ReadOnlyFSp (fstype))
    WorkingDirSafetyCheck (workingdir);

  Dstr codeset;
  GetCodeset (codeset, fstype);

  if ((fstype == iso || fstype == isoj) && force_perms)
    fprintf (stderr, "Warning:  -p has no effect for iso or isoj.\n");
  if ((fstype != ext2) && (fstype != le2a) && (fstype != le2f) && gurue2)
    fprintf (stderr, "Warning:  -l only works on ext2/le2a/le2f.\n");
  if ((fstype != le2a) && (fstype != le2f) && nornd)
    fprintf (stderr, "Warning:  -nornd applies only to le2a/le2f.\n");

  const bool compressSizes = (fstype == squashfs);
  if (compressSizes)
    printf ("Hashing source files and getting compressed sizes\n");
  else
    printf ("Hashing source files\n");
  FileList flist, rlist;
  for (++argi; argi<argc; ++argi)
    addSomething (argv[argi], flist, true, compressSizes);
  for (std::vector<Dstr>::iterator it (replicants.begin());
       it != replicants.end();
       ++it)
    addSomething (*it, rlist, true, compressSizes);

  // Purging duplicates here means they got hashed twice already, but it
  // avoids a lot of problems with aliasing.
  fixDuplicates (flist, rlist);

  double ratio = 1.0;
  unsigned imgserialnum=0;
  while (!flist.empty()) {
    char tempfilename[20];
    sprintf (tempfilename, "image%03d.iso", ++imgserialnum);
    Dstr imgfilename (imagedir);
    imgfilename += tempfilename;
    if (ReadOnlyFSp (fstype))
      buildReadOnlyImage (imgfilename, workingdir, flist, rlist, discsize,
			  ratio, fstype, force_perms, codeset);
    else
      buildWritableImage (imgfilename, workingdir, flist, rlist, discsize,
			  fstype, force_perms, gurue2, nornd, codeset);
    if ((!flist.empty()) && pausewait)
      wait_for_user();
  }

  return 0;
}
